#include "synth.h"

void usage(char *prog)
{
	fprintf(stderr,"\n");
	fprintf(stderr,"Usage: %s [ -f <filename> ] [ -p ] [ -e ] [ -t ] [ -s ]"
                                                             " [ -i ]\n",prog);
	fprintf(stderr,"       [ -C <cachesize> ] [ -R <ringbuffersize> ]"
                         " [ -S <scheduling cycle size> ]");
    fprintf(stderr,"       [ -D <debug> ]\n\n");
	fprintf(stderr,"   -f   the .arts-file to load\n");
	fprintf(stderr,"   -p   publish the specified file on the server\n");
	fprintf(stderr,"   -e   execute the specified file on the server\n");
	fprintf(stderr,"   -t   free synthserver after usage when possible\n");
	fprintf(stderr,"   -i   retrieve status information of the server\n");
	fprintf(stderr,"   -s   show (realtime) settings\n");
	fprintf(stderr,"   -R   set ringbuffer size im samples, the argument MUST be a 2^n value\n");
	fprintf(stderr,"   -S   set scheduling cycle size in samples\n");
	fprintf(stderr,"   -C   set cache size in KB\n");
	fprintf(stderr,"   -D   set debugging (1 to enable, 0 to disableable)\n");
	exit(1);
}

main(int argc, char **argv)
{
	CORBA::ORB_var orb = CORBA::ORB_init( argc, argv, "mico-local-orb" );
	CORBA::BOA_var boa = orb->BOA_init( argc, argv, "mico-local-boa" );

	CORBA::Object_var obj = orb->bind ("IDL:Arts/Synthesizer:1.0",
												"inet:localhost:8888");
	Arts::Synthesizer_var Synthesizer = Arts::Synthesizer::_narrow (obj);

	if(CORBA::is_nil(Synthesizer))
	{
		fprintf(stderr,"%s will only work when the Arts Server is running\n",
					argv[0]);
		exit(1);
	}
	char filename[255];
	bool publish = false, execute = false,fileok = false,terminate = false, showrt = false, changert = false, info = false;
	int optch;
	Arts::RealtimeConfig rtc = Synthesizer->RTConfig();

	while((optch = getopt(argc,argv,"tf:piesS:C:R:D:")) > 0)
	{
		switch(optch)
		{
			case 'f': strncpy(filename,optarg,255);
					  fileok = true;
				break;
			case 'p': publish = true;
				break;
			case 'e': execute = true;
				break;
			case 't': terminate = true;
				break;
			case 'i': info = true;
				break;
			case 's': showrt = true;
				break;
			case 'C': rtc.CacheSizeKB = atol(optarg);
					  changert = true;
				break;
			case 'R': rtc.RingBufferSize = atol(optarg);
					  changert = true;
				break;
			case 'S': rtc.Cycles = atol(optarg);
					  changert = true;
				break;
			case 'D': rtc.Debug = atol(optarg);
					  changert = true;
				break;
			default: usage(argv[0]);
				break;
		}
	}

	if(changert) showrt = true;

	if(!fileok && !showrt && !info) usage(argv[0]);

	if(changert)
	{
		Synthesizer->RTConfig(rtc);
		printf("Realtime settings updated.\n");
		rtc = Synthesizer->RTConfig();
	}

	if(showrt)
	{
		char debugstatus[10] = "disabled";
		if(rtc.Debug) strcpy(debugstatus,"enabled");
		printf("\n");
		printf("    Scheduling cycle size: %ld samples (delay %2.2f ms with 44.1 kHz)\n",rtc.Cycles,(float)rtc.Cycles/44.1);
		printf("         Ring buffer size: %ld samples\n",rtc.RingBufferSize);
		printf("Cache size (for WAVs etc): %ld KB\n",rtc.CacheSizeKB);
		printf("    Debugging information: %s\n",debugstatus);
		printf("\n");
	}

	if(info)
	{
		Arts::Status status = Synthesizer->getStatus();
		printf("\n");
		printf("   Active modules: %ld\n",status.module_count);
		printf("        CPU usage: %2.2f%%\n",status.cpu_usage*100);
		printf("\n");
		Arts::ResourceInfoSeq *resources = Synthesizer->Resources();
		unsigned long i;
		for(i=0;i<resources->length();i++)
		{
			Arts::ResourceInfo &ri = (*resources)[i];
			printf("%10ld %s\n",ri.Used,(const char *)ri.Description);
		}
		if(i>0) printf("\n");
	}

	if(!fileok) exit(0);

    FILE *infile = fopen(filename,"r");
	if(!infile)
	{
		fprintf(stderr,"%s: can't open file '%s'!\n",argv[0],filename);
		exit(EXIT_FAILURE);
	}

    Arts::StringSeq_var strseq = new Arts::StringSeq;
 
    char line[1024];
    unsigned long i = 0;
 
    while(fgets(line,1024,infile))
    {
        // cut eventual CR and/or LFs at the end of the line
        while(strlen(line) && line[strlen(line)-1] < 14)
            line[strlen(line)-1] = 0;
 
        strseq->length(i+1);
        (*strseq)[i++] = CORBA::string_dup(line);
    }
    fclose(infile);

	// this is ugly, since the synthesizer may terminate if not referenced
	// so this option should probably make room for a more well behaved one?

	if(terminate)
		Synthesizer->incRef();

	Arts::StructureDesc_var StructureDesc = Synthesizer->createStructureDesc(); 
	StructureDesc->incRef();
    StructureDesc->loadFromList(strseq);

	if(publish)
		Synthesizer->publishStructureDesc(StructureDesc);

	if(execute)
	{
		Arts::ArtsServerSeq preferredservers;	// oh well, don't care where
												// our stuff gets executed
	    // just in case synthesis has been halted before,
   		// restart it and hope we'll have enough computing power now
    	Synthesizer->Reset();
    	long execID = Synthesizer->createStructure(StructureDesc,
				preferredservers);
    	assert(execID);
		printf(" <press return to end execution> \n\n"); 
		getchar();
    	Synthesizer->freeStructure(execID);
	}
	StructureDesc->decRef();

	if(terminate)
		Synthesizer->decRef();
}
