	/*

	Copyright (C) 1998 Stefan Westerfeld
                       stefan@space.twc.de

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    */

#include "synthmodule.h"
#include "synth_impl.h"
#include "debug.h"

class BusClient
{
public:
	virtual void
		configureBus(list<BusClient *>& clientlist, unsigned long channels)
		{ /* only in downlinks */ }
	virtual SynthModule *Module() = 0;
};


static class BusManager
{
	struct Bus
	{
		string name;
		list<BusClient *> clients;
		list<BusClient *> servers;
	};
	list<Bus *> BusList;

public:
	Bus* findBus(string name);
	void reBuild(Bus *bus);
	void erase(BusClient *busclient);
	void addClient(string busname, BusClient *client);
	void removeClient(BusClient *client);
	void addServer(string busname, BusClient *server);
	void removeServer(BusClient *server);
	static BusManager *getBusManager();
}	*the_BusManager = 0;

BusManager *BusManager::getBusManager()
{
	if(!the_BusManager) the_BusManager = new BusManager;
	return(the_BusManager);
}

BusManager::Bus *BusManager::findBus(string name)
{
	list<Bus *>::iterator bi;

	for(bi = BusList.begin(); bi != BusList.end(); bi++)
	{
		if((*bi)->name == name) return(*bi);
	}
	Bus *bus = new Bus;
	bus->name = name;
	BusList.push_back(bus);
	return(bus);
}

void BusManager::reBuild(Bus *bus)
{
	list<BusClient *>::iterator client,server;
	long channels = 2;

	artsdebug("rebuilding bus %s\n",bus->name.c_str());
	artsdebug(" - %ld channels\n",channels);
	artsdebug(" - %d clients\n",bus->clients.size());
	artsdebug(" - %d servers\n",bus->servers.size());

	for(client = bus->clients.begin(); client != bus->clients.end(); client++)
	{
		artsdebug(" => client %s\n",(*client)->Module()->getClassName());
	}

	for(server = bus->servers.begin(); server != bus->servers.end(); server++)
	{
		artsdebug(" => server %s\n",(*server)->Module()->getClassName());
		(*server)->configureBus(bus->clients, channels);
	}
}

void BusManager::erase(BusClient *busclient)
{
	list<Bus *>::iterator bi;

	for(bi = BusList.begin(); bi != BusList.end(); bi++)
	{
		list<BusClient *>::iterator client,server;
		int hits = 0;
		Bus *bus = *bi;

	for(client = bus->clients.begin(); client != bus->clients.end(); client++)
		if((*client) == busclient)
		{
			bus->clients.erase(client);
			client = bus->clients.begin();
			hits++;
		}

	for(server = bus->servers.begin(); server != bus->servers.end(); server++)
		if((*server) == busclient)
		{
			bus->clients.erase(server);
			server = bus->servers.begin();
			hits++;
		}
		if(hits != 0) reBuild(bus);
	}
}

void BusManager::addClient(string busname, BusClient *client)
{
	Bus *bus = findBus(busname);
	bus->clients.push_back(client);
	reBuild(bus);
}

void BusManager::removeClient(BusClient *client)
{
	erase(client);
}

void BusManager::addServer(string busname, BusClient *server)
{
	Bus *bus = findBus(busname);
	bus->servers.push_back(server);
	reBuild(bus);
}

void BusManager::removeServer(BusClient *server)
{
	erase(server);
}

class Synth_BUS_UPLINK :public SynthModule,BusClient
{
	const int LEFT = 0;
	const int RIGHT = 1;
	const int BUSNAME = 2;
	BusManager *bm;
public:
	void Initialize(string info);
	void DeInitialize();
	void Calculate() { /* dont calculate */ }
	void CalculateBlock(unsigned long samples) { /* dont calculate */ }
	string getParams() { return("left,right,_busname;"); }
	static void *Creator() { return new Synth_BUS_UPLINK; };
	SynthModule *Module();
};

void Synth_BUS_UPLINK::Initialize(string info)
{
	bm = BusManager::getBusManager();
	bm->addClient(info, this);
	haveCalculateBlock = true;
}

void Synth_BUS_UPLINK::DeInitialize()
{
	bm->removeClient(this);
}

SynthModule *Synth_BUS_UPLINK::Module()
{
	return(this);
}

ModuleClient MC_Synth_BUS_UPLINK(SynthModule::get_MS,"Synth_BUS_UPLINK",Synth_BUS_UPLINK::Creator);

class Synth_BUS_DOWNLINK :public SynthModule,BusClient
{
	const int CLIENTS = 0;
	const int BUSNAME = 1;
	const int LEFT = 0;
	const int RIGHT = 1;
	BusManager *bm;

	unsigned long bus_clients;
	unsigned long bus_channels;

public:
	void Initialize(string info);
	void DeInitialize();
	void Calculate() { /* only have CalculateBlock */ };
	void CalculateBlock(unsigned long samples);
	string getParams() { return("clients,_busname;left,right"); }
	static void *Creator() { return new Synth_BUS_DOWNLINK; };
	void configureBus(list<BusClient *>& clientlist, unsigned long channels);
	SynthModule *Module();
};

void Synth_BUS_DOWNLINK::Initialize(string info)
{
	bm = BusManager::getBusManager();
	bm->addServer(info, this);
	haveCalculateBlock = true;
}

void Synth_BUS_DOWNLINK::DeInitialize()
{
	Synthesizer->removeDynamicConnections(this);
	bm->removeServer(this);
}

SynthModule *Synth_BUS_DOWNLINK::Module()
{
	return(this);
}

void Synth_BUS_DOWNLINK::configureBus(list<BusClient *>& clientlist, unsigned long channels)
{
	Synthesizer->removeDynamicConnections(this);

	list<BusClient *>::iterator i;
	for(i=clientlist.begin();i != clientlist.end();i++)
	{
		BusClient *client = *i;
		SynthModule *smod = client->Module();
		artsdebug("1 client is: %s\n",smod->getClassName());
		Synthesizer->addDynamicConnection(this,smod,channels);
	}
	bus_channels = channels;
	bus_clients = clientlist.size();
}

// sum up the data coming from the bus

void Synth_BUS_DOWNLINK::CalculateBlock(unsigned long samples)
{
	float *sig;

	unsigned long channel, client;

	for(channel = 0; channel < bus_channels; channel++)
	{
		float *out_start = out[channel], *out_end = out[channel]+samples;

		sig = out_start;
		while(sig < out_end) *sig++ = 0.0;		// first zero the fragment

		for(client=0; client<bus_clients; client++)
		{
			float *input=in[1 + client*bus_channels + channel]; // 1+ is ugly ;)

			assert(inConnCount >= 1 + client*bus_channels + channel);
			sig = out_start;
			while(sig < out_end) *sig++ += *input++;
											// add input data from client
		}

		float *gain = in[CLIENTS];
		sig = out_start;
		while(sig < out_end) *sig++ /= *gain++;  // scale down data
	}
}

ModuleClient MC_Synth_BUS_DOWNLINK(SynthModule::get_MS,"Synth_BUS_DOWNLINK",Synth_BUS_DOWNLINK::Creator);
