    /*

    Copyright (C) 1998 Stefan Westerfeld
                       stefan@space.twc.de

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    */

#include "midi.h"
#include "interfaces.h"
#include "synth_impl.h"
#include "debug.h"
#include <math.h>
#include <list>

void MidiChannel_impl::noteOn( CORBA::Octet channel, CORBA::Octet note,
														CORBA::Octet volume )
{
	list<MidiReceiver *>::iterator i;
	for(i=receivers.begin();i != receivers.end();i++)
	{
		(*i)->noteOn(channel,note,volume);
	}
}

void MidiChannel_impl::noteOff( CORBA::Octet channel, CORBA::Octet note )
{
	list<MidiReceiver *>::iterator i;
	for(i=receivers.begin();i != receivers.end();i++)
		(*i)->noteOff(channel,note);
}

void MidiChannel_impl::subscribe(MidiReceiver *receiver)
{
	receivers.push_back(receiver);
}

void MidiChannel_impl::unSubscribe(MidiReceiver *receiver)
{
	receivers.remove(receiver);
}

//---------------------------------------------------------------------------

class MidiEvent :public Interface_MIDI_NOTE
{
	int _channel, _note, _velocity;
	float outfrequency, outvelocity, outpressed;
	bool init;
public:

	long StructureID;

	int channel();
	int note();
	void off();
	void Initialize(string info);
	void CalculateBlock(unsigned long samples);
	MidiEvent(int channel, int note, int velocity);
};

int MidiEvent::note()
{
	return _note;
}

int MidiEvent::channel()
{
	return _channel;
}

void MidiEvent::off()
{
	if(init)
	{
		outpressed = 0;
	}
	else
	{
		printf("MidiEvent: Humm? Got noteoff before being initialized?\n");
	}
}

MidiEvent::MidiEvent(int channel, int note, int velocity)
{
	_velocity = velocity;
	_note = note;
	_channel = channel;
	init = false;
}

void MidiEvent::Initialize(string info)
{
/*
    float freq[]  = { 261.7,277.2,293.7,311.2,329.7,349.3,370.0,392.0,
					  415.3,440.0, 466.2,493.9, 0 };
    float zhoch[24];
	int i,octave = _note/12;

	for(i=0;i<24;i++) zhoch[i] = (1 << i);
	//outfrequency = freq[_note%12];
	outfrequency *= zhoch[_note/12] / zhoch[4];
*/
	outfrequency = 261.63 * pow(2,(float)_note/12)/32.0;
	outpressed = 1;
	outvelocity = (float)_velocity/127.0;
	init = true;
	haveCalculateBlock = true;
	artsdebug("pitch: %d, frequency: %2.2f\n",_note,outfrequency);
}

void MidiEvent::CalculateBlock(unsigned long samples)
{
	float *sigfrequency = out[FREQUENCY], *sigpressed = out[PRESSED],
          *sigvelocity = out[VELOCITY], *sigend = sigfrequency + samples;

	while(sigfrequency < sigend)
	{
		*sigfrequency++ = outfrequency;
		*sigpressed++   = outpressed;
		*sigvelocity++  = outvelocity;
	}	
}

class Synth_MIDI_ROUTER :public SynthModule, MidiReceiver
{
protected:
	Arts::StructureDesc_var structure;

	string structurename;
	int mychannel,maxvoices;
	int i;
	float count;

	long OldStructChangeCount;

	list<MidiEvent *> activeEvents;

	const int CHANNEL = 0;
	const int MAXVOICES = 1;
	const int COUNT = 0;

public:
	void cleanUp();
	void noteOn(int channel, int note, int velocity);
	void noteOff(int channel, int note);
	void Initialize(string info);
	void CalculateBlock(unsigned long samples);
	void DeInitialize();
	void Calculate();
	string getParams() { return("channel,maxvoices,_structure;count"); }
	static void *Creator() { return new Synth_MIDI_ROUTER; };
};

ModuleClient MC_Synth_MIDI_ROUTER(SynthModule::get_MS,"Synth_MIDI_ROUTER",
												Synth_MIDI_ROUTER::Creator);

void Synth_MIDI_ROUTER::Initialize(string info)
{
	OldStructChangeCount = 0;
	structurename = info;
	count = 0;
	maxvoices = 0;
	i = 0;
	Synthesizer->getMidiChannel()->subscribe(this);
	haveCalculateBlock = true;

	structure = 0;

	Arts::StructureDesc_var sd =
			Synthesizer->lookupStructureDesc(structurename.c_str());
	if(!sd)
	{
		fprintf(stderr,"FATAL!!-> can't find structure %s for midi synth\n",
							structurename.c_str());
	}
	else
	{
		sd->incRef();
		structure = Synthesizer->expandStructureDesc(sd);
		structure->incRef();
		sd->decRef();
	}
}

void Synth_MIDI_ROUTER::DeInitialize()
{
	if(structure)
	{
		structure->decRef();
	}
	Synthesizer->getMidiChannel()->unSubscribe(this);

	// this will kill the remaining notes that are still playing the
	// hard way (just free their structures) - otherwise, it may happen
	// that some of them remain playing forever, since they never have
	// a chance to see the midi-release event without their MIDI_ROUTER
	list<MidiEvent *>::iterator i;
	
	for(i=activeEvents.begin();i != activeEvents.end();i++)
	{
		MidiEvent *ev = (*i);
		if(Synthesizer->isExecuting(ev->StructureID))
			Synthesizer->freeStructure(ev->StructureID);
	}
}

void Synth_MIDI_ROUTER::Calculate()
{
}

void Synth_MIDI_ROUTER::CalculateBlock(unsigned long samples)
{
	i+=samples;
	if(i>500)
	{
		if(Synthesizer->StructChangeCount() != OldStructChangeCount)
			cleanUp();

		OldStructChangeCount = Synthesizer->StructChangeCount();
		i = 0;
	}
	mychannel = (int)(in[CHANNEL][samples-1] + 0.5);
	maxvoices = (int)(in[MAXVOICES][samples-1] + 0.5);

	float *outcount = out[COUNT], *outend = &out[COUNT][samples];

	while(outcount < outend) *outcount++ = count;
}

void Synth_MIDI_ROUTER::noteOn(int channel, int note, int velocity)
{
	// simply cut voices that are over the given maximum
	// could certainly be done more elegantly

	if(mychannel == channel && (count < maxvoices))
	{
		MidiEvent *event = new MidiEvent(channel,note,velocity);
		event->setClassName("Interface_MIDI_NOTE");

		activeEvents.push_back(event);

		list<SynthModule *> interfaces;
		interfaces.push_back(event);
/*
		Arts::StructureDesc_var sd =
			Synthesizer->lookupStructureDesc(structurename.c_str());
		if(!sd)
		{
			fprintf(stderr,"FATAL!!-> can't find structure %s for midi synth\n",
								structurename.c_str());
		}
		else
		{
*/
		if(structure)
		{
			Arts::ArtsServerSeq preferredservers;
			event->StructureID =
				Synthesizer->createConnectedStructure(structure,
								preferredservers, interfaces);
		}
/*
		}
*/
		count = activeEvents.size();
	}
}

void Synth_MIDI_ROUTER::noteOff(int channel, int note)
{
	list<MidiEvent *>::iterator i;
	
	for(i=activeEvents.begin();i != activeEvents.end();i++)
	{
		MidiEvent *ev = (*i);
		if(ev->channel() == channel && ev->note() == note) ev->off();
	}
}

void Synth_MIDI_ROUTER::cleanUp()
{
	list<MidiEvent *>::iterator i;
	
	for(i=activeEvents.begin();i != activeEvents.end();i++)
	{
		MidiEvent *ev = (*i);
		if(!Synthesizer->isExecuting(ev->StructureID))
		{
			artsdebug("purged a module\n");
			activeEvents.erase(i);
			//remove(ev);
			//delete ev;  FIXME?
			i = activeEvents.begin();
			count = activeEvents.size();
		}
	}
}
