	/*

	Copyright (C) 1998 Stefan Westerfeld
                       stefan@space.twc.de

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    */

#include <sys/types.h>
#include <sys/time.h>
#include <math.h>
#include <stdio.h>
#include <ctype.h>
#include <strings.h>
#include <unistd.h>
#include <stdlib.h>
#include <malloc.h>
#include <time.h>
#include "synth_impl.h"
#include "resources.h"
#include "midi.h"
#include "midibus.h"
#include "structures.h"
#include "ioevent.h"
#include "synth_server.h"
#include "debug.h"
#include "execrequest.h"
#include "receiver_impl.h"
 
extern "C" {
#include "sound.h"
#include "utils.h"
}

#define Die(a) { fprintf(stderr,"ERROR: %s\n",a); exit(1); }

//****************** Synthesizer *******************

Synthesizer_impl::Synthesizer_impl()
{
	Arts::RealtimeConfig InitRTConfig;

	describe("Synthesizer");

	InitRTConfig.Cycles = 128;		// calculate 128 cycles at once by default
	InitRTConfig.CacheSizeKB = 8192;	// 8 megs of cache are reasonable
	InitRTConfig.RingBufferSize = 64;	// ringbuffer size (2^n)
	InitRTConfig.Debug = false;		// no debugging by default

	RTConfig(InitRTConfig);

	artsdebug("... Synthesizer_impl created\n");
	SynthModules = NULL;
	_Cache = new Cache;
	MCount = 0;
	_StructChangeCount = 0;
	MS_SynthModule = (ModuleServer<SynthModule> *)SynthModule::get_MS();
	nextID = 1;
	nextMID = 1;

	inScheduler = false;

	_MidiChannel = new MidiChannel_impl();
	_ModuleBroker = new ModuleBroker_impl();

	printf("%d\n",_ModuleBroker->_refcnt());

	_ModuleBroker->incRef();

	initModules();

	_StructureExpander = new StructureExpander(_ModuleBroker,this);
}

void Synthesizer_impl::cleanUp()
{
	unsigned long i;

	// release published structures when being shut down
	for(i=0;i<_Structures.length();i++)
		_Structures[i]->decRef();

	// remove published modules from broker

	_ModuleBroker->serverShutdown(this);
	_ModuleBroker->decRef();

	// delete StructureExpander, which will free it's references to
	// Synthesizer & ModuleBroker
	delete _StructureExpander;
}

MidiChannel_impl *Synthesizer_impl::getMidiChannel()
{
	return _MidiChannel;
}

long Synthesizer_impl::StructChangeCount()
{
	return _StructChangeCount;
}

Synthesizer_impl::~Synthesizer_impl()
{
	CORBA::release(_MidiChannel);
	artsdebug("... Synthesizer_impl destroyed\n");
}

Arts::ModuleBroker *Synthesizer_impl::moduleBroker()
{
	assert(_ModuleBroker);
	return(Arts::ModuleBroker::_duplicate(_ModuleBroker));
}

void Synthesizer_impl::RTConfig(const Arts::RealtimeConfig &newconfig)
{
	assert(newconfig.Cycles > 0);
	_RTConfig = newconfig;
	setartsdebug(_RTConfig.Debug);
}

Arts::RealtimeConfig Synthesizer_impl::RTConfig()
{
	return _RTConfig;
}

Cache *Synthesizer_impl::getCache()
{
	return _Cache;
}

bool Synthesizer_impl::Exec()
{
/** old style (one sample) scheduling **
	long i,cycles = _RTConfig.Cycles;
	while(cycles--)
	{
		for(i=0;i<MCount;i++) SynthModules[i]->Calculate();
	}
****************************************/

/** new style (dynamic size) scheduling **/

	long cycles = _RTConfig.Cycles;
	unsigned long *done = calloc(MCount,sizeof(unsigned long));
	long i,incomplete, died = 0;

	//printf("entering; cycles = %d\n",cycles);
	do {
		incomplete = 0;		/* assume we have calculated all cycles for all
								consumers, and later increment if some are
								still missing */
		for(i=0;i<MCount;i++)
		{
			if(SynthModules[i]->clients == 0 && SynthModules[i]->enabled)
			{
				//printf("consumber = %s,done = %d, cycles = %d\n",SynthModules[i]->getClassName(),done[i],cycles);
				/* a module whose input is not comsumed from other modules
					is a "push delivery" style module, such as speakers,
					or writing audio to log file, etc. and has to get
					external requests from the scheduling system */

				if(done[i] != cycles)
					done[i] += SynthModules[i]->request(cycles-done[i]);
				assert(done[i] <= cycles);
				if(done[i] != cycles) incomplete++;
				//printf("*scheduler*\n");
				died ++;
				if(died > 10000)
				{
					free(done);
					return false;
				}
			}
		}
	} while(incomplete);

	//printf("<=> done!!\n");
	free(done);
	return true;
}

void Synthesizer_impl::Run()
{
	inScheduler = true;
	bool eresult = Exec();
	inScheduler = false;

	if(!eresult)
	{
		status_halted = true;
		fprintf(stderr,"seems to be uncalculatable?\n");
	}

	while(killList.size())
	{
		long mid = *killList.begin();
		// FIXME: this accesses the execution server variables from the flow
		// server, which is ugly
		long id = 0;
		list<ExecRequest *>::iterator ei;
		for(ei=execRequests.begin();ei != execRequests.end(); ei++)
		{
			ExecRequest *er = *ei;
			if(er->ServerID() == ID() && er->MID() == mid)
				id = er->StructureID();
		}
		if(id == 0)
			artsdebug("kill mid %ld failed\n",mid);
		else
			freeStructure(id);
		killList.erase(killList.begin());
	}

	if(status_halted)
	{
		usleep(10000);
		return;
	}

	//fprintf(stderr,"Using %ld modules for synthesis\n",MCount);

	// Calculate cpu usage.
	status_cycles += _RTConfig.Cycles;
	if(status_cycles > 20000)
	{
		// do cache cleanups periodically
		status_cache_usage = _Cache->cleanUp(_RTConfig.CacheSizeKB*1024);

	    float cpu_time = (clock()-status_clocks)/(float)CLOCKS_PER_SEC;
		float real_time = benchmark(BENCH_END);

		if(cpu_time > 0 && real_time > 0) // there may be wraparounds
		{
			status_cpu_usage = cpu_time / real_time;
			if(status_cpu_usage > 0.9)	  // more than 90%  -> not good!
			{
				status_stalled ++;
			}
			else status_stalled = 0;

			if(status_stalled > 5)
			{
				// ok, cancel synthesis due to cpu overload!
				status_halted = true;
			}
		}

		// prepare for next checkpoint
		status_clocks = clock(); 
		status_cycles -= 20000;
		benchmark(BENCH_BEGIN);
	}
}

Arts::Status Synthesizer_impl::getStatus()
{
	Arts::Status status;

	status.cpu_usage = status_cpu_usage;
	status.cache_usage = status_cache_usage;
	status.halted = status_halted;
	status.module_count = MCount;
	return(status);
}

void Synthesizer_impl::Reset()
{
	status_halted = false;
	status_stalled = 0;
	status_cpu_usage = 0;
	status_clocks = clock();
	status_cycles = 0;	
	benchmark(BENCH_BEGIN);
}

CORBA::Long Synthesizer_impl::getModuleCount()
{
	assert(0);
	return(MS_SynthModule->getModuleCount());
}

CORBA::Boolean Synthesizer_impl::getModuleInfo( CORBA::Long nr, Arts::ModuleInfo*& info )
{
	assert(0);
}

void collectPorts( Arts::PortTypeSeq& Ports, Arts::StringSeq& PortNames,
     SynthModule *module, SynthModule::ArgType mdir, Arts::PortDirection pdir )
{
	string portname;
	char *pn;
	long portcount = 0;
	long len;

	Arts::PortType porttype;

	while((portname = module->getArgument(portcount,mdir)) != "")
	{
		porttype.Direction = pdir;
		porttype.DataType = Arts::audio_data;
		porttype.ConnType = Arts::stream;

		if(portname[0] == '_')		// is there a better way?
		{
			pn = strdup(portname.c_str());
			pn ++;					// cut the first byte
			portname = pn;
			pn --;					// but free the right thing
			free(pn);

			porttype.DataType = Arts::string_data;
			porttype.ConnType = Arts::property;
		}

		artsdebug("#%d: %s\n",portcount,portname.c_str());

		len = Ports.length();
		Ports.length(len+1);
		Ports[len] = porttype;

		len = PortNames.length();
		PortNames.length(len+1);
		PortNames[len] = (const char *)portname.c_str();

		portcount++;
	}
}

// FlowServer

void Synthesizer_impl::initModules()
{
	long nr;

	_ID = _ModuleBroker->mkID();

	for(nr=0;nr < MS_SynthModule->getModuleCount(); nr++)
	{
		char *name=MS_SynthModule->getModuleName(nr);
		if(name)
		{
			SynthModule *m = (SynthModule *)MS_SynthModule->getModule(name);

			if(m)
			{
				Arts::ModuleInfo info;
				info.name = CORBA::string_dup(name);
				info.isInterface = (strncmp(name,"Interface",9) == 0);
				info.isStructure = false;

				collectPorts(info.ports,info.portnames,m,
								SynthModule::arg_in, Arts::input);
				collectPorts(info.ports,info.portnames,m,
								SynthModule::arg_out, Arts::output);

				_ModuleBroker->publishModule(this,info);

				delete m;
			}
		}
	}
}

Arts::StructureDesc *Synthesizer_impl::expandStructureDesc(Arts::StructureDesc
															*structuredesc)
{
	return _StructureExpander->expandStructure(structuredesc);
}

// Factory for StructureDesc
Arts::StructureDesc *Synthesizer_impl::createStructureDesc()
{
	artsdebug("creating new structure desc\n");

	Arts::StructureDesc_ptr sd =
		new StructureDesc_impl(Arts::ModuleBroker::_duplicate(_ModuleBroker));
	return Arts::StructureDesc::_duplicate(sd);
}

// Execution Manager
CORBA::Long Synthesizer_impl::createStructure1(Arts::StructureDesc *sd)
{
	// this is just there to test some mico-2.2.6 behaviour, which I suppose
	// is a bug:
	// passing a structuredesc from outside (artsbuilder) to this server
	// causes the referencecount to be modified (so calling this method ten
	// times increments the reference count by ten)
	printf("refcnt %d\n",sd->_refcnt());
	return 6;
}

CORBA::Long Synthesizer_impl::createStructure(Arts::StructureDesc
					*structuredesc, const Arts::ArtsServerSeq& preferredservers)
{
	// no interfaces to the outside world
	//
	// this is just for using the same code for creating structures
	// that are connected with interfaces to the outside world and
	// creating self contained structures
	list<SynthModule *> interfaces;

	return createConnectedStructure(structuredesc,preferredservers,interfaces);
}


// Execution Manager
long Synthesizer_impl::createConnectedStructure(Arts::StructureDesc
			*structuredesc, const Arts::ArtsServerSeq& preferredservers,
			list<SynthModule*> interfaces)
{
	assert(!inScheduler);

	_StructChangeCount++;

	// first step: create modules, allocate output buffers

	// since "interfaces" are not yet specified in the CORBA interfaces,
	// they are passed to the FlowServer beside CORBA, which is an ugly
	// hack

	_global_interfaces_hack = interfaces;

	Arts::ModuleDescSeq_var modules;
	Arts::StructureDesc_var expandedsd = 0;
	if(!structuredesc->containsStructures())
	{
		artsdebug("Execute 0: no expansion necessary, no substructures used\n");
 		modules = structuredesc->Modules();
	}
	else
	{
		artsdebug("Execute 0 <Expansion>\n");
		expandedsd = expandStructureDesc(structuredesc);
		expandedsd->incRef();
 		modules = expandedsd->Modules();
	}

	list<ExecRequest *> newexlist;
	list<ExecRequest *>::iterator ei;
	ExecRequest *er;

	artsdebug("Execute I <Creation>\n");
	unsigned long m;

	for(m=0;m<modules->length();m++)
	{
		Arts::ModuleDesc *md = (*modules)[m];
		CORBA::String_var modname = md->Name();
		Arts::ArtsServerSeq_var servers =
						_ModuleBroker->whichServerProvides(modname);
		artsdebug("checking whether module %s is present\n",
					(const char *)modname);
		assert(servers->length());

		// FIXME: might do some nicer server selection algorithm
		long serverID = -1;
		Arts::ArtsServer *server = 0;

		if(servers->length() > 1)
		{
			artsdebug("more that one possible server choice for module %s\n",
								(const char *)modname);
			unsigned long i,s;
			for(i=0;i<preferredservers.length() && !server;i++)
			{
				long goodID = preferredservers[i]->ID();
				for(s=0;s<servers->length() && !server;s++)
				{
					if((*servers)[s]->ID() == goodID)
					{
						artsdebug("selected server with ID %d\n",goodID);
						server = (*servers)[s];
					}
				}
			}
			if(!server)
				artsdebug("WARNING: no server due to preferences possible\n");
		}
		if(!server)		// default choice - take the first server that is there
			server = (*servers)[0];

		serverID = server->ID();

		er = 0;

		for(ei=newexlist.begin();!er && ei != newexlist.end();ei++)
		{
			ExecRequest *thiser = *ei;
			if(thiser->ServerID() == serverID)
			{
				// good, we've already got some server doing work
				// for us, just put the module in there
				er = thiser;
			}
		}

		// server not yet connected to do the creation for us; connect
		if(!er)
		{
			er = new ExecRequest(server,nextID);
			newexlist.push_back(er);
			execRequests.push_back(er);
		}
		er->createModule(md);
	}

	// second step: establish local connections between the modules
	for(ei=newexlist.begin();ei != newexlist.end();ei++)
		(*ei)->localConnectModules();

	// remote connection setup:
	for(ei=newexlist.begin();ei != newexlist.end();ei++)
	{
		list<ExecRequest *>::iterator ei2;
		for(ei2=newexlist.begin();ei2 != newexlist.end();ei2++)
		{
			if(ei != ei2)
				(*ei)->remoteConnectModules(*ei2);
		}
	}

	// finalize the modules (does the last initialization before start)

	artsdebug("finalization:\n");
	for(ei=newexlist.begin();ei != newexlist.end();ei++)
		(*ei)->finalizeModules();

	// finally start the modules
	artsdebug("start:\n");
	for(ei=newexlist.begin();ei != newexlist.end();ei++)
		(*ei)->startModules();

	if(expandedsd)
	{
		artsdebug("Free expanded structure\n");
		expandedsd->decRef();
	}

	artsdebug("Execute ok\n");

	return nextID++;
}

// Dynamic Connections? Execution Manager?
void Synthesizer_impl::addDynamicConnection (SynthModule *target, SynthModule *source, unsigned long size)
{
	assert(!inScheduler);
	unsigned long i;
	artsdebug(">>>>>>>>>> ADC: target is %s source is %s\n",target->getClassName(),source->getClassName());
	artsdebug(">>>>>>>>>> ADC: source->inConnCount = %d\n",source->inConnCount);
	artsdebug(">>>>>>>>>> ADC: target->inConnCount = %d\n",target->inConnCount);
	artsdebug(">>>>>>>>>> ADC: add channels: %d\n",size);
	assert(source->inConnCount >= size);
	for(i=0;i < size;i++)
	{
		SynthConnection *newconn = new SynthConnection(target,0);
		newconn->isDynamic = true;
		target->assignInConn(newconn);
		// FIXME: will fail when constant stuff
		newconn->connectToSource(source->inConn[i]->sourceconn);
	}
	artsdebug(">>>>>>>>>> after target->inConnCount = %d\n",target->inConnCount);
}

// Dynamic Connections? Execution Manager?
void Synthesizer_impl::removeDynamicConnections (SynthModule *module)
{
	assert(!inScheduler);

	unsigned long i,remove=0;

	for(i=0;i<module->inConnCount;i++)
	{
		if(module->inConn[i]->isDynamic)
		{
			remove++;
			if(module->inConn[i]->sourceconn)
				module->inConn[i]->disconnectFromSource(module->inConn[i]->sourceconn);
			delete module->inConn[i];
		}
		else
		{
			// dynamic channels should ONLY be attached at the end!
			assert(remove == 0);
		}
	}
	module->inConnCount -= remove;

	module->inConn =
		realloc(module->inConn,sizeof(SynthConnection *)*module->inConnCount);
	// FIXME: free memory here (?)
}

// Flow Server
void Synthesizer_impl::kill(long id)
{
	// Is called by Synth_STRUCT_KILL
	//
	// As freeing while a scheduling cycle is done is not possible, we push
	// that request to a list and execute the kills after scheduling

	killList.push_back(id);
}

// Execution Manager
CORBA::Boolean Synthesizer_impl::isExecuting(CORBA::Long id)
{
	list<ExecRequest *>::iterator ei;
	for(ei=execRequests.begin(); ei != execRequests.end(); ei++)
	{
		if((*ei)->StructureID() == id) return true;
	}
	return false;
}

// Execution Manager + Flow Server
CORBA::Boolean Synthesizer_impl::freeStructure(CORBA::Long id)
{
	assert(!inScheduler);

	artsdebug("freeing structure id = %ld\n",id);

	// Should notify the parent that the structure it is executing
	// no longer exists. Currently there is only polling to check that.

	_StructChangeCount++;

	// kill all those modules that are executed on some server out there
	// and belong to that structure

	list<ExecRequest *>::iterator ei,nextei;
	for(ei = execRequests.begin(); ei != execRequests.end(); ei = nextei)
	{
		ExecRequest *er = *ei;
		nextei = ei;
		++nextei;
		if(er->StructureID() == id)
		{
			er->deleteModules();
			execRequests.erase(ei);
			delete er;
		}
	}

	return true; // FIXME: return code
}

// Structure Repository
void Synthesizer_impl::publishStructureDesc(Arts::StructureDesc *structuredesc)
{
	Arts::StructureDesc_ptr structure;

	unsigned long l = 0;
	bool found = false;

	// create deep copy by saving and reloading
	Arts::StringSeq_var strings = structuredesc->saveToList();

	// reload below

	while(l < _Structures.length() && !found)
	{
		CORBA::String_var name1=_Structures[l]->Name();
		CORBA::String_var name2=structuredesc->Name();

		found = !strcmp(name1,name2);
		if(!found) l++;
	}

	if(!found) // didn't find a structure with the same name
	{
		_Structures.length(l+1);

		structure = createStructureDesc();
		_Structures[l] = structure;
		structure->incRef();
	}
	else		// if we did, just overwrite this one
	{
		structure = _Structures[l];
		structure->clear();		// clear old contents
	}
	structure->loadFromList(strings);

	Arts::ModuleInfo_var mi = structure->ExternalInterface();
	_ModuleBroker->publishModule(this,*mi);
}

// Structure Repository
Arts::StructureDesc *Synthesizer_impl::lookupStructureDesc(const char *name)
{
	unsigned long i;

	for(i=0;i<_Structures.length();i++)
	{
		CORBA::String_var aname = _Structures[i]->Name();
		if(strcmp(aname,name) == 0)
			return Arts::StructureDesc::_duplicate(_Structures[i]);
	}

	return 0;
}

// Structure Repository
Arts::StructureDescSeq *Synthesizer_impl::publishedStructures()
{
	return(new Arts::StructureDescSeq(_Structures));
}

// Flow Server?
Arts::ResourceInfoSeq *Synthesizer_impl::Resources()
{
	Arts::ResourceInfoSeq *resources = new Arts::ResourceInfoSeq;
	list<ResourceMonitor *> &rlist = *ResourceList::the()->getResources();
	list<ResourceMonitor *>::iterator i;
	unsigned long nr = 0;

	resources->length(rlist.size());
	for(i=rlist.begin();i != rlist.end();i++)
	{
		Arts::ResourceInfo& ri = (*resources)[nr++];
		ri.Used = (*i)->Used();
		ri.Description = CORBA::string_dup((*i)->Description());
	}
	return(resources);
}

/************ Flow System Server *************/

CORBA::Long Synthesizer_impl::createModules()
{
	return nextMID++;
}

CORBA::Boolean Synthesizer_impl::createModule(CORBA::Long mid,
												Arts::ModuleDesc *desc)
{
	unsigned long p;

	CORBA::String_var modname = desc->Name();
	Arts::PortDescSeq_var ports = desc->Ports();
	long RBSize = _RTConfig.RingBufferSize;
	artsdebug("creating module %s\n",(char *)modname);
	
	SynthModule *NewModule;

	// since "interfaces" are not yet specified in the CORBA interfaces,
	// they are passed to the FlowServer beside CORBA, which is an ugly
	// hack
	list<SynthModule *>& interfaces = _global_interfaces_hack;
	
	if(desc->isInterface())
	{
		list<SynthModule *>::iterator iface = interfaces.begin();
		bool found = false;
		artsdebug(" => setting up interface connection\n");
	
		while(!found && iface != interfaces.end())
		{
			artsdebug("[%s]<=>[%s]\n",(*iface)->getClassName(),(char *)modname);
			if(strcmp((*iface)->getClassName(),modname) == 0)
			{
				found = true;
				NewModule = (*iface);

				// only use once of course!
				interfaces.remove(NewModule);
			}
			iface++;
		}
		assert(found);
	}	
	else
	{
		NewModule = (SynthModule *)MS_SynthModule->getModule(modname);
	}
	assert(NewModule);
	NewModule->Synthesizer = this;
	
	NewModule->ID = desc->ID();
	NewModule->mID = mid;
	NewModule->textargs = "";

/*	FIXME: may use some other solution here
	if(!desc->isInterface())
	{
		...?
	}
*/
	//NewModule->enabled = !desc->isInterface();

	//NewModule->enabled = true;  !!! WARNING: don't enable here, as mico
	// may try to do flow system evaluations in between

	SynthModules = (SynthModule **)realloc(SynthModules,
									sizeof(SynthModule *)*(MCount+1));
	SynthModules[MCount++] = NewModule;

	for(p=0;p<ports->length();p++)
	{
		Arts::PortDesc *pd = (*ports)[p];
		Arts::PortType pt = pd->Type();

		if(pt.Direction == Arts::input)
		{
			assert(pt.ConnType != Arts::event); // no event channels yet

			if(pt.ConnType == Arts::stream)
			{
				assert(pt.DataType == Arts::audio_data);
				
				SynthConnection *conn =
					new SynthConnection(NewModule,pd->ID());
				NewModule->assignInConn(conn);

				if(pd->isConnected())
				{
					Arts::PortDescSeq_var conns = pd->Connections();
					assert(conns->length() == 1);
					Arts::PortDesc *pd2 = (*conns)[0];

					long wantedID = pd2->ID();
					long foundnr = 0;
					conn->needConnect = true;
					conn->wantedID = wantedID;
				}
				else
				{
					float initialvalue = 0.0;

					if(pd->hasValue())
						initialvalue = pd->FloatValue();

					conn->allocBuffer(initialvalue,RBSize);
				}
			}
			if(pt.ConnType == Arts::property)
			{
				assert(pt.DataType == Arts::string_data);
				assert(NewModule->textargs == "");

				// only one textarg may be passed in that primitive form
				CORBA::String_var arg = pd->StringValue();
				NewModule->textargs = arg;
				artsdebug("got stringproperty; %s\n",
					NewModule->textargs.c_str());
			}
		}
		if(pt.Direction == Arts::output)
		{
			artsdebug("creating port id %ld\n",pd->ID());

			SynthConnection *conn =
				new SynthConnection(NewModule,pd->ID());
			conn->allocBuffer(0.0,RBSize);
			NewModule->assignOutConn(conn);
		}
	}
}

CORBA::Boolean Synthesizer_impl::remoteConnectModules(CORBA::Long mid,
		CORBA::Long remotemid, const Arts::ModuleDescSeq& modules,
		Arts::ArtsServer *server)
{
	map<long,bool> portMap;

	artsdebug("remoteConnectModules to server %d called\n",server->ID());
	unsigned long m,i;

	for(m=0;m<modules.length();m++)
	{
		Arts::ModuleDesc *md = modules[m];
		Arts::PortDescSeq_var ports = md->Ports();

		unsigned long p;
		for(p=0;p<(*ports).length();p++)
		{
			portMap[(*ports)[p]->ID()] = true;		// he has that port
		}
	}
	for(i=0;i<MCount;i++)
	{
		SynthModule *NewModule = SynthModules[i];

		if(NewModule->mID == mid)
		{
			long p;
			for(p=0;p<NewModule->inConnCount;p++)
			{
				SynthConnection *conn = NewModule->inConn[p];
				if(conn->needConnect)
				{
					artsdebug("needconnect for %s\n",NewModule->getClassName());
					if(portMap[conn->wantedID])  // yes, it's on that server
					{
						artsdebug("found a source for %s\n",
								NewModule->getClassName());
						artsdebug("requesting from server\n");
						Arts::Receiver *rcv = new Receiver_impl(conn);
						server->requestSignal(remotemid,conn->wantedID, rcv);
						conn->needConnect = false;
					}
				}
			}
		}
	}	
}

CORBA::Boolean Synthesizer_impl::localConnectModules(CORBA::Long mid)
{
	if(!SynthModules) return;
	
	artsdebug("Execute II localConnectModules\n");
	long i;

	map<long,SynthConnection *> portMap;

	for(i=0;i<MCount;i++)
	{
		SynthModule *NewModule = SynthModules[i];

		if(NewModule->mID == mid)
		{
			long p;
			for(p=0;p<NewModule->outConnCount;p++)
			{
				portMap[NewModule->outConn[p]->ID] = NewModule->outConn[p];
			}
		}
	}
	for(i=0;i<MCount;i++)
	{
		SynthModule *NewModule = SynthModules[i];

		if(NewModule->mID == mid)
		{
			long p;
			for(p=0;p<NewModule->inConnCount;p++)
			{
				SynthConnection *conn = NewModule->inConn[p];
				if(conn->needConnect)
				{
					SynthConnection *source = portMap[conn->wantedID];
					if(!source)
					{
						artsdebug("NewModule %s missing source (remote?)\n",
							NewModule->getClassName());
					}
					else
					{
						conn->connectToSource(source);
						conn->needConnect = false;
					}
				}
			}
		}
	}
}

CORBA::Boolean Synthesizer_impl::finalizeModules(CORBA::Long mid)
{
	if(!SynthModules) return;

	artsdebug("Execute III <Prepare & Initialization> in finalizeModules\n");
	long i;

	for(i=0;i<MCount;i++)
	{
		SynthModule *NewModule = SynthModules[i];

		if(NewModule->mID == mid)
		{
			artsdebug("Initialize: textargs = %s\n",
						NewModule->textargs.c_str());
			artsdebug("NewModule %s characteristics: in %ld, out %ld\n",
						NewModule->getClassName(),NewModule->inConnCount,
						NewModule->outConnCount);
			NewModule->Initialize(NewModule->textargs);
	
			// EXPERIMENTAL: prepareExecution after Initialize
			NewModule->prepareExecution();
			NewModule->enabled = true;
		}
	}
}

void Synthesizer_impl::startModules(CORBA::Long mid)
{
	if(!SynthModules) return;

	artsdebug("Execute IV <IO Initialization> in startModules\n");
	long i;

	for(i=0;i<MCount;i++)
	{
		SynthModule *NewModule = SynthModules[i];
		if(NewModule->mID == mid)
		{
			int infd,outfd;
			ArtsIOEvent *ev;
	
			NewModule->getfds(infd,outfd);
	
			if(infd != -1)
			{
				ev = new ArtsIOEvent(this);
				ArtsOrb->dispatcher()->rd_event(ev,infd);
				ev->mID = NewModule->mID;
				IOEvents.push_back(ev);
			}
			if(outfd != -1)
			{
				ev = new ArtsIOEvent(this);
				ArtsOrb->dispatcher()->wr_event(ev,outfd);
				ev->mID = NewModule->mID;
				IOEvents.push_back(ev);
			}
		}
	}
}

void Synthesizer_impl::deleteModules(CORBA::Long mid)
{
	long NCount = 0,freed = 0,i;
	artsdebug("executing deletemodules %ld\n",mid);

	list<ArtsIOEvent *>::iterator ioevent;
	for (ioevent=IOEvents.begin(); ioevent != IOEvents.end(); ioevent++)
	{
		ArtsIOEvent *ev = (*ioevent);
		if(ev->mID == mid)
		{
			ArtsOrb->dispatcher()->remove(ev,CORBA::Dispatcher::All);
			delete ev;
			IOEvents.erase(ioevent);

			// and restart from the beginning because our iterator could
			// now point to nirwana
			ioevent = IOEvents.begin();

			artsdebug("removed io event\n");
		}
	}

	if(SynthModules)
	{
		// kill those that are local and match the id

		// 1st step: deinitialize them
		for(i=0;i<MCount;i++)
		{
			if(SynthModules[i]->mID == mid)
				SynthModules[i]->DeInitialize();
		}

		// 2nd step: delete them
		for(i=0;i<MCount;i++)
		{
			if(SynthModules[i]->mID == mid)
			{
				delete SynthModules[i];
				freed++;
			}
			else SynthModules[NCount++] = SynthModules[i]; // "compress" modules
		}

		MCount = NCount;
		SynthModules = (SynthModule **)realloc(SynthModules,
									sizeof(SynthModule *)*MCount);
		if(MCount == 0) SynthModules = 0;

		if(freed)
		{
			artsdebug("freed %ld modules\n",freed);
		}
	}
	artsdebug("deletemodules ok\n");
}

CORBA::Long Synthesizer_impl::ID()
{
	return _ID;
}

void Synthesizer_impl::requestSignal(CORBA::Long mid, CORBA::Long portid,
										Arts::Receiver *receiver)
{
}

// old connection code
#if 0
		artsdebug("checking whether module %s is local\n",(const char *)modname);

		isLocal[m] = false;
		unsigned long i;
		for(i=0;i < servers->length();i++)
		{
			if((*servers)[i]->ID() == ID())
			{
				isLocal[m] = true;
			}
		}

		if(isLocal[m]) artsdebug("module %s is local\n",(const char *)modname);
		         else  artsdebug("module %s is remote\n",(const char *)modname);
		if(isLocal[m])
		{
			Arts::PortDescSeq_var ports = md->Ports();
			artsdebug("creating module %ld %s\n",m,(char *)modname);
	
			SynthModule *NewModule;
	
			if(md->isInterface())
			{
				list<SynthModule *>::iterator iface = interfaces.begin();
				bool found = false;
				artsdebug(" => setting up interface connection\n");
	
				while(!found && iface != interfaces.end())
				{
					artsdebug("[%s] <=> [%s]\n",(*iface)->getClassName(),(char *)modname);
					if(strcmp((*iface)->getClassName(),modname) == 0)
					{
						found = true;
						NewModule = (*iface);
	
						// only use once of course!
						interfaces.remove(NewModule);
					}
					iface++;
				}
				assert(found);
			}	
			else
			{
				NewModule = (SynthModule *)MS_SynthModule->getModule(modname);
			}
			assert(NewModule);
			NewModule->Synthesizer = this;
	
			NewModules[m] = NewModule;
			NewModule->ID = md->ID();
			NewModule->executionID = nextID;
			NewModule->textargs = "";
	
			if(!md->isInterface())
			{
				SynthModules = (SynthModule **)realloc(SynthModules,
										sizeof(SynthModule *)*(MCount+1));
				SynthModules[MCount++] = NewModule;
			}

			for(p=0;p<ports->length();p++)
			{
				Arts::PortType pt;
	
				pd = (*ports)[p];
				pt = pd->Type();
	
				if(pt.Direction == Arts::input)
				{
					assert(pt.ConnType != Arts::event); // no event channels yet
	
					if(pt.ConnType == Arts::stream)
					{
						assert(pt.DataType == Arts::audio_data);
						
						SynthConnection *conn =
							new SynthConnection(NewModule,pd->ID());
						NewModule->assignInConn(conn);
	
						if(pd->isConnected())
						{
							Arts::PortDescSeq_var conns = pd->Connections();
							assert(conns->length() == 1);
							Arts::PortDesc *pd2 = (*conns)[0];
	
							long wantedID = pd2->ID();
							long foundnr = 0;
							conn->needConnect = true;
							conn->wantedID = wantedID;
						}
						else
						{
							float initialvalue = 0.0;
	
							if(pd->hasValue())
								initialvalue = pd->FloatValue();
	
							conn->allocBuffer(initialvalue,RBSize);
						}
					}
					if(pt.ConnType == Arts::property)
					{
						assert(pt.DataType == Arts::string_data);
						assert(NewModule->textargs == "");
	
						// only one textarg may be passed in that primitive form
						CORBA::String_var arg = pd->StringValue();
						NewModule->textargs = arg;
						artsdebug("got stringproperty; %s\n",
							NewModule->textargs.c_str());
					}
				}
				if(pt.Direction == Arts::output)
				{
					artsdebug("creating port id %ld\n",pd->ID());
	
					SynthConnection *conn =
						new SynthConnection(NewModule,pd->ID());
					conn->allocBuffer(0.0,RBSize);
					NewModule->assignOutConn(conn);
				}
			}
		}
#endif
