	/*

	Copyright (C) 1998 Stefan Westerfeld
                       stefan@space.twc.de

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    */

#include "synthmodule.h"
#include "resources.h"
#include "debug.h"

static long SynthConnection::instances = 0;
ResourceMonitor RM_SynthConnection_instances(&SynthConnection::instances,"SynthConnection object instances");

static long SynthBuffer::instances = 0;
static ResourceMonitor RM_SynthBuffer_instances(&SynthBuffer::instances,"SynthBuffer object instances");

static long SynthBuffer::allbytes = 0;
static ResourceMonitor RM_SynthBuffer_allbytes(&SynthBuffer::allbytes,"Total bytes in all SynthBuffers");

string SynthModule::getArgument(int nr, ArgType direction)
{
	string result, arglist = getParams();
	int wanted = nr, have = 0;
	unsigned long i;

	if(direction == arg_out)
		wanted += 512;	// second string of args behind ';'
						// more than 512 arguments currently unsupported ;)

	for (i=0;i<arglist.length();i++)
	{
		switch(arglist[i])
		{
			case ',': have++;
				break;
			case ';': have += 512 - (have & 511);
				break;
			default:  if(have == wanted) result += arglist[i];
				break;
		}
	}
	return(result);
};

void SynthModule::ProcessData(string data)
{
	fprintf(stderr,"data could not be processed by %s\n",getName().c_str());
}

void SynthModule::DeInitialize()
{
	// do nothing
}

void SynthModule::getfds(int &infd, int &outfd)
{
	// if we are an io performing module (midi event processing, sound output,
    // ..., we are likely to open out files here and return them to the
    // synthesizer, which will perform scheduling if io is to do

	// setting fds to -1 means, we don't have them

    infd = -1;   // no input performed by this module
	outfd = -1;  // no output performed by this module
}

void *SynthModule::get_MS()
{
	static ModuleServer<SynthModule> *MS = 0;
	if(!MS)
	{
		MS = new ModuleServer<SynthModule>;
#ifndef MODULESERVER_DEBUG
	}
#else
		artsdebug("Creating a ModuleServer for SynthModules\n");
	}
	else
	{
		artsdebug("Obtaining a ModuleServer reference for SynthModules\n");
	}
	artsdebug("ModuleServer is at %lx\n",MS);
#endif

	return(MS);
}

SynthModule::SynthModule()
{
	inConn = 0;
	outConn = 0;

	in = out = 0;		// for fast access from Calculate(Block) to ringbuffer data

	inConnCount = 0;
	outConnCount = 0;

	clients = 0;

	enabled = false;

	haveCalculateBlock = false;	// default is currently still single sample
}

SynthModule::~SynthModule()
{
	unsigned long i;
	artsdebug("inConnCount = %d, outConnCount = %d\n",inConnCount,outConnCount);
	for(i = 0;i<inConnCount;i++) delete inConn[i];
	for(i = 0;i<outConnCount;i++) delete outConn[i];
	if(inConnCount)
	{
		free(inConn);
		free(in);
	}
	if(outConnCount)
	{
		free(outConn);
		free(out);
	}
}

void SynthModule::assignInConn(SynthConnection *newInConn)
{
	inConn = realloc(inConn, sizeof(SynthConnection *) * (inConnCount+1));
	inConn[inConnCount] = newInConn;

	inConnCount++;

	in = realloc(in, sizeof(float *) * inConnCount);
}

void SynthModule::assignOutConn(SynthConnection *newOutConn)
{
	outConn = realloc(outConn, sizeof(SynthConnection *) * (outConnCount+1));
	outConn[outConnCount] = newOutConn;

	outConnCount++;

	out = realloc(out, sizeof(float *) * outConnCount);
}

void SynthModule::prepareExecution()
{
	unsigned long i;

	// reset scheduling information
    NeedCycles = 0;
    CanPerform = 0;
    Busy = 0;
    BusyHit = 0;
 
	clients = 0;

	for(i=0;i<outConnCount;i++) clients += outConn[i]->destcount;

	unsigned long preprod = 1024;
	for(i=0;i<outConnCount;i++)
	{
		unsigned long room = outConn[i]->outRoom();
		if(preprod > room) preprod = room;
		assert(outConn[i]->buffer->position == 0);
		out[i] = outConn[i]->buffer->data;
	}

	preprod = PreProduce(preprod);
	if(preprod)
	{
		for(i=0;i<outConnCount;i++)
		{
			outConn[i]->buffer->position += preprod;
			outConn[i]->buffer->needread += preprod * outConn[i]->destcount;
		}
	}
}

/* request a module to calculate a number of turns
   will return -1 if busy, count that has been done otherwise */

long SynthModule::request(long amount)
{
	int in,have_in,need_more,have_done,have_done_total = 0;

	if(Busy) { BusyHit++; return(-1); }

	Busy = 1;
	if(NeedCycles < amount)
	{
		NeedCycles = amount;
	}

	//artsdebug("DSP %s: request of %d cycles\n",getClassName(),amount);
	do
	{
		/* assume we can satisfy the request */
		CanPerform = NeedCycles;

		/* then, check wether the input channels supply enough data to do so. */

		for(in=0;in<inConnCount;in++)
		{
			have_in = inConn[in]->haveIn();

			if(have_in < NeedCycles)
			{
				//artsdebug("connection %d indicates to have %d, "
                //       "thats not enough\n", in,have_in);
				/* if we can't calculate enough stuff due to a certain
				   ingoing connection, go to the associated module and
				   tell it that we need more data
				*/
				need_more = NeedCycles - have_in;
				//artsdebug("requesting %d\n", need_more);
				/* when there is no source (constant input value), then
					we don't need to request something, it's just that
					it can't supply more because the buffer isn't big enough
				*/
				if(inConn[in]->sourcemodule)
					inConn[in]->sourcemodule->request(need_more);

				have_in = inConn[in]->haveIn();
	
				//artsdebug("now connection %d indicates to have %d\n",in,have_in);
				if(CanPerform > have_in) CanPerform = have_in;
			}
		}
		have_done = calc(CanPerform);
		have_done_total += have_done;

		/*
		if(dsp->m_BusyHit != 0) artsdebug("got busyhit: %s\n",dsp->m_Name);
		*/

	} while(BusyHit && NeedCycles != CanPerform && have_done);
	/* makes only sense to repeat if
		 - there was a busyhit which indicates we are in a feedback loop
		 - we should have done more than we have done
         - actually something could be calculated
     */
	Busy = 0;
	return(have_done);
}

/* This routine would now actually let the plugin calculate some data
   that means generate sinus waves, mix audio signals etc.

   The number of cycles is guaranteed to work without input underrun
   by the flow system. But the routine still needs to check output
   stall situations (ring buffer full).
*/

long SynthModule::calc(long cycles)
{
	int i,room;

	/* output sanity check:
	   when there is not enough room in one of the buffers, we
	   can't calculate that much cycles
	*/
	for(i=0;i<outConnCount;i++)
	{
		room = outConn[i]->outRoom();
		if(room < cycles)
		{
			cycles = room;
			/*
			artsdebug("- reduced calculation to %d due to lack of space\n",cycles);
			*/
		}
	}

	if(cycles == 0) return(0);

	//artsdebug("DSP %s: calculation of %d cycles\n",getClassName(),cycles);
	/* input sanity check:
		it's guaranteed that we have enough input, but do the check
		anyway... - you never know
	*/
	for(i=0;i<inConnCount;i++)
	{
		/* otherwise input is "overconsumed" */
		assert(inConn[i]->haveIn() >= cycles);

		/* check sanity of the needread setting:
			either data comes from fixed value (sourcemodule not assigned)
			then we can't expect needread to contain sensible setting,
			but otherwise: needread should have expected that we'll want
			to read the data and thus be more than the cycles! */

		assert((!inConn[i]->sourcemodule)
				|| (inConn[i]->buffer->needread >= cycles));
	}

	if(haveCalculateBlock)
	{
		int donecycles = 0,cando = 0,j;
	
		while(donecycles != cycles)
		{	
			cando = cycles-donecycles;

			for(j=0;j<inConnCount;j++)
			{
				int rbsize = inConn[j]->buffer->size;
				int rbmask = rbsize - 1;
				int rbpos = (inConn[j]->position+donecycles) & rbmask;
				in[j] = &inConn[j]->buffer->data[rbpos];
				if(cando > (rbsize-rbpos)) cando = rbsize - rbpos;
			}

			for(j=0;j<outConnCount;j++)
			{
				int rbsize = outConn[j]->buffer->size;
				int rbmask = rbsize - 1;
				int rbpos = (outConn[j]->buffer->position+donecycles) & rbmask;
				out[j] = &outConn[j]->buffer->data[rbpos];
				if(cando > (rbsize-rbpos)) cando = rbsize - rbpos;
			}
	
			CalculateBlock(cando);
			donecycles += cando;
		}
		assert(donecycles == cycles);
	}
	else
	{
		for(i=0;i<cycles;i++)
		{
			int j;

			for(j=0;j<inConnCount;j++)
			{
				int rbm = inConn[j]->buffer->size-1;	// assume size is 2^n
				in[j] = &inConn[j]->buffer->data[(inConn[j]->position+i) & rbm];
			}

			for(j=0;j<outConnCount;j++)
			{
				int rbm = outConn[j]->buffer->size-1;	// assume size is 2^n
				out[j] =
			&outConn[j]->buffer->data[(outConn[j]->buffer->position+i) & rbm];
			}

			Calculate();
		}
	}

	// actually update buffer by subtracting consumed input
	for(i=0;i<inConnCount;i++)
	{
		inConn[i]->position += cycles;				/* consumes input */
		inConn[i]->buffer->needread -= cycles;		/* consumes input */
	}

	// and adding fresh output
	for(i=0;i<outConnCount;i++)
	{
		outConn[i]->buffer->position += cycles;
		outConn[i]->buffer->needread += cycles * outConn[i]->destcount;
	}

	NeedCycles -= cycles;
	CanPerform -= cycles;
	return(cycles);
}

void SynthModule::CalculateBlock(unsigned long size)
{
	assert(false);	// never do this ;)
}

unsigned long SynthModule::PreProduce(unsigned long maxsamples)
{
	// default: no preproduction possible
	return(0);
}

// SynthBuffer...:

SynthBuffer::SynthBuffer(float initialvalue, unsigned long size)
{
	this->size = size;
	data = malloc(sizeof(float) * size);

	setValue(initialvalue);

	position = 0;
	needread = 0;
	instances++;
	allbytes += size;
}

void SynthBuffer::setValue(float value)
{
	unsigned long i;
	for(i=0;i<size;i++) data[i] = value;
}

SynthBuffer::~SynthBuffer()
{
	instances--;
	allbytes -= size;
	free(data);
}

// SynthConnection...:

SynthConnection::SynthConnection(SynthModule *module, long ID)
{
	this->ID = ID;
	this->module = module;
	position = 0;
	destcount = 0;
	destmodule = 0;
	sourcemodule = 0;
	buffer = 0;
	sourceconn = 0;
	isDynamic = false;
	needConnect = false;
	freeBufferOnExit = false;
	instances++;
}

SynthConnection::~SynthConnection()
{
	if(destcount) free(destmodule);
	if(freeBufferOnExit)
	{
		assert(buffer);
		delete buffer;
	}
	instances--;
}

void SynthConnection::connectToSource(SynthConnection *source)
{
	sourceconn = source;
    buffer = source->buffer;
    sourcemodule = source->module;
 
    source->destmodule = realloc(source->destmodule,
								sizeof(SynthModule *)*(source->destcount+1));
    source->destmodule[source->destcount] = module;
	source->destcount++;
	artsdebug("ConnectToSource: %s->destcount set to %d\n",source->module->getClassName(),source->destcount);

	position = buffer->position;
}

void SynthConnection::disconnectFromSource(SynthConnection *source)
{
	SynthModule **ndestmodule = 0;
	unsigned long i,ndestcount = 0;

	artsdebug("DISCONNECT -> recalculating %s destcount: before %ld\n",source->module->getClassName(),source->destcount);
	for(i=0;i<source->destcount;i++)
	{
		if(source->destmodule[i] != module)
		{
    		ndestmodule = realloc(ndestmodule, sizeof(SynthModule *)*(ndestcount+1));
    		ndestmodule[ndestcount++] = source->destmodule[i];
		}
	}
	if(source->destcount) free(source->destmodule);
	source->destmodule = ndestmodule;
	source->destcount = ndestcount;
	artsdebug("DISCONNECT -> recalculating %s destcount: now %ld\n",source->module->getClassName(),source->destcount);
}

void SynthConnection::allocBuffer(float initialvalue, unsigned long size)
{
	assert(!buffer);
	assert(!freeBufferOnExit);
	buffer = new SynthBuffer(initialvalue,size);
	freeBufferOnExit = true;
}
